{******************************************************************************}
{* This code was written by : Petr Grebenkin                                  *}
{* with big thanks to programmers Inprise Corporation and Jamie Hart(TXBase)  *}
{******************************************************************************}
{* TFosDSet is a shreware component for Delphi 5 which gives access           *}
{* to dBase, FoxPro data files without needing the BDE.                       *}
{* Version 2.0                                                                *}
{******************************************************************************}
{* Users of the TFoxDSet component must accept the following disclaimer of    *}
{* warranty:                                                                  *}
{*                                                                            *}
{* TFoxDSet is supplied as is. The authors disclaims all warranties, expressed*}
{* or implied, including, without limitation, the warranties of               *}
{* merchantability and of fitness for any purpose. The authors assume no      *}
{* liability for damages, direct or consequential, which may result from the  *}
{* use of TFoxDSet.                                                           *}
{******************************************************************************}
unit FoxDSet;

{$R-,T-,H+,X+}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
   Db, DbConsts , jpeg;

Const
  dB3Memo = $83;
  dB4Memo = $84;
  FoxMemo = $F5;
  FoxV = $30;

type
  PRecInfo = ^TRecInfo;
  TRecInfo = packed record
    RecordNumber: Longint;
    UpdateStatus: TUpdateStatus;
    BookmarkFlag: TBookmarkFlag;
  end;

    {Type DBF}
    TTypeDBF = (Fox30,Fox2x,dBxMemo,dB5,dB4,dB3,UnKnown);
  THeader = packed record
    VersionNumber: Byte;
    LastUpdated: Array[1..3] Of Byte;
    NumRecords: DWORD;
    HeaderLen: Word;
    RecordLen: Word;
    Reserved: Word;
    IncTrans: Byte;
    EncFlag: Byte;
    Reserved2: Array[1..12] Of Byte;
    MdxFlag: Byte;
    LangDriver: Byte;
    Reserved3: Word;
  End;

Type
  TDBTHeader = packed record
    NextBlock: DWORD;
    BlockSize: DWORD;
    Reserved: Array[1..504] Of Char;
  End;

Type
  TFieldStruct = packed Record
    FieldName: Array[1..11] Of Char;
    FieldType: Char;
    Address: DWORD;
    FieldLength: Byte;
    Decimals: Byte;
    Reserved: Word;
    WorkArea: Byte;
    Reserved2: Word;
    SetFields: Byte;
    Reserved3: Array[1..7] Of Byte;
    IndexFlag: Byte;
  End;

    TUconvEvent = procedure(Src, Dest: PChar; ToOem: Boolean) of object;
    TUCodeEvent = procedure(buffer:Pointer; count:integer; Mode: TBlobStreamMode) of object;

    {Mode CopyTo}
    TModeCopyTo = (cAll, cCurrentRec);

    RFilter = record
    val: Variant;
    IsPole:boolean;
    offset:integer;
    count: Integer;
    end;

  TFoxDSet = class(TDataSet)
  private
    FOnConvert: TUconvEvent;
    FOnEncode: TUCodeEvent;
//    FLocale: TLocale;
    FMasterLink: TMasterDataLink;
    FMasterOffcet: integer;
    FMasterLen: integer;
    FOnMaster: Boolean;
{user}
    FReadOnly:Boolean;
    FTableName:string;
    FIsOpen:Boolean;
    FTypeDBF: Byte;
    DBFile: TFileStream;
    MemoFile: TFileStream;
    FHeader:THeader;
    FDBTHeader: TDBTHeader;
    FNFields:integer;
    FieldStruct: Array[0..128] Of TFieldStruct;
    FCurrentRecordF: DWord;
    FRecordSize: Word;
    FBookmarkOfs: Word;
    FRecInfoOfs: Word;
    FBlobCacheOfs: Word;
    FRecBufSize: Word;
    FBufferForFilter:Pointer;
    FShowDelRec:boolean;
    FFilterData: array of RFilter;
    FFilterDataCount: integer;
    FIndexName: string;
    FIndexOffcet: Integer;
    FIndexLen: integer;
    FTableCreate: Boolean;
    FWorkDir: String;
    FTranlateOn: boolean;
     Procedure FillHeaderInfo;
     procedure InitBufferPointers(GetProps: Boolean);
     function CheckFilter(buffer:PChar):boolean;
     function GreatLess( var i:integer; F:boolean):boolean;
     function AlignString( var i:integer; F:boolean):boolean;
    function GetIndexFieldNames: string;
    procedure SetIndexFieldNames(const Value: string);
    procedure MasterChanged(Sender: TObject);
    procedure MasterDisabled(Sender: TObject);
    function Links(Buffer: PChar):boolean;
    function GetTableType: TTypeDBF;
    procedure SetTableType(Value: TTypeDBF);
   protected
//    procedure SetLocale(Value: TLocale);
     procedure DataConvert(Field: TField; Source, Dest: Pointer; ToNative: Boolean); overload; override;
    procedure InternalInsert;  overload; override;
  { abstract methods, overload }
    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure SetFilterText(const Value: string); override;
    procedure SetFiltered(Value: Boolean); override;
    {user}
    function GetActiveRecBuf(var RecBuf: PChar): Boolean;
    Function ReadARecord(RecordBuffer:PChar):integer;
    Procedure WriteARecord(RecordBuffer:PChar);
    function ope(i:integer):boolean;
    procedure SetTableName(Name:string);
    procedure SetReadOnly(Value: Boolean);
    Procedure UpdateFieldData(FieldNo: integer; RecBuf: Pointer; Buffer: Pointer);
    function GetTableCreate: Boolean;
    procedure SetTableCreate( Value: Boolean);
//    procedure SetFieldData(Field: TField; Buffer: Pointer; NativeFormat: Boolean); overload; override;
    {for programming}
    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    function GetDataSource: TDataSource; override;
    procedure SetDataSource(Value: TDataSource);
    function GetMasterFields: string;
    procedure SetMasterFields(const Value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; overload; override;
    function GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean; overload; override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer; override;
    procedure SetShowDelRec(Value: Boolean);
    procedure CreateTable;
    procedure CopyTo(Dest: TFoxDSet);
    procedure CopyToFile(Name: string);
    procedure AppendFrom(Source: TFoxDSet);
    procedure AppendFromFile( Name: string);
    procedure Pack;
    Procedure MoveRecs(source,tagert:integer);
    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant; override;
    function Translate(Src, Dest: PChar; ToOem: Boolean): Integer;  override;
    function GetFieldS(FieldD: TFieldDef; var Fields:TFieldStruct):boolean;
    function SizeDataFile:integer;
    property WorkDir: String read FWorkDir;
  published
    property Active;
    property AutoCalcFields;
    property FieldDefs;
    property TableName:string read FTableName write SetTableName;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly stored True default False;
    property ShowDelRec: Boolean read FShowDelRec write SetShowDelRec stored True default False;
    property IndexFieldNames: string read GetIndexFieldNames write SetIndexFieldNames;
    property MasterSource: TDataSource read GetDataSource write SetDataSource;
    property MasterFields: string read GetMasterFields write SetMasterFields;
    property TableType: TTypeDBF read GetTableType write SetTabletype;
    property TableCreate: Boolean read GetTableCreate write SetTableCreate;
    property Filter;
    property Filtered;
    property FilterOptions;
{Events from TDataSet}
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property BeforeRefresh;
    property AfterRefresh;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
    property OnUserConvert: TUconvEvent read FOnConvert write FOnConvert;
    property OnUserEncode: TUCodeEvent read FOnEncode write FOnEncode;
 end;
procedure Register;
type
  TMemoStream = class(TStream)
  private
    FDataSet: TFoxDSet;
    FField: TField;
    nlen:cardinal;
    StartPos: integer;
    FMode: TBlobStreamMode;
    FSize: Longint;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;
 { TMemoFieldP }
 type
  TMemoFieldP = class(TBlobField)
  private
    TypeGraph:DWORD;
    FieldKnow: boolean;
    Procedure SaveToPicture(Source: TPersistent);
    Procedure LoadFromPicture(Source: TPersistent);
  protected
    procedure AssignTo(Dest: TPersistent); override;
  public
    procedure Assign(Source: TPersistent); override;
    procedure Clear; override;
    constructor Create(AOwner: TComponent); override;
  published
    property Transliterate default True;
  end;

implementation

procedure invertDWORD(var w:DWORD);
type invert= packed record
     case byte of
     0: (orig: DWORD);
     1: (   m: array[0..3] of byte);
end;
var buf:^invert;
    bufB:byte;
begin
   buf:=@w;
   bufB:=buf.m[0]; buf.m[0]:=buf.m[3]; buf.m[3]:=bufB;
   bufB:=buf.m[1]; buf.m[1]:=buf.m[2]; buf.m[2]:=bufB;
end;


procedure Register;
begin
  RegisterComponents('Data Access', [TFoxDSet]);
end;

constructor TFoxDSet.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMasterLink := TMasterDataLink.Create(Self);
  FMasterLink.OnMasterChange := MasterChanged;
  FMasterLink.OnMasterDisable := MasterDisabled;
  DecimalSeparator:='.';
    DBFile := nil;
    MemoFile :=nil;
    FOnConvert := nil;
    FOnEncode := nil;
    FTranlateOn :=True;
    FFilterDataCount := 0;
//  NestedDataSetClass := TNestedTable;
end;

destructor TFoxDSet.Destroy;
begin
  FMasterLink.Free;
  inherited Destroy;
end;

{Procedure TFoxDSet.SetLocale(Value: TLocale);
Begin
  FLocale := Value;
End; }
{buffers}
Function TFoxDSet.AllocRecordBuffer: PChar;
    Begin
    if FIsOpen then
    Result := AllocMem(FRecBufSize)
    else
    Result := nil;
    End;

Procedure TFoxDSet.FreeRecordBuffer(var Buffer: PChar);
    Begin
    freemem(buffer);
    End;
{bookmark}
Procedure TFoxDSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
    Begin
    Move(Buffer[FBookmarkOfs], Data^, BookmarkSize);
    End;

Procedure TFoxDSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
    Begin
    if Data <> nil then
    Move(Data^, Buffer[FBookmarkOfs], BookmarkSize);
    End;

function TFoxDSet.CompareBookmarks(Bookmark1, Bookmark2: TBookmark): Integer;
const
  RetCodes: array[Boolean, Boolean] of ShortInt = ((2,-1),(1,0));
begin
  { Check for uninitialized bookmarks }
  Result := RetCodes[Bookmark1 = nil, Bookmark2 = nil];
  if Result = 2 then
  Result := PInteger(Bookmark1)^ - PInteger(Bookmark2)^;
end;

Function TFoxDSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
    Begin
    Result := PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
    End;

Procedure TFoxDSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
    Begin
    PRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
    End;

{record}
Function TFoxDSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var Status: integer;
    Accept: Boolean;
    WorkRec: DWORD;
Begin
If FIsOpen then Begin
        Result := grOK;
repeat
  WorkRec := FCurrentRecordF; {Status as TMP}
   case GetMode of
     gmCurrent:;
     gmNext: inc(WorkRec);
     gmPrior: if WorkRec>0 then dec(WorkRec);
   end;
  if ( WorkRec > FHeader.NumRecords ) then begin  Result := grEOF; exit; end;
  if ( WorkRec < 1 ) then begin  Result := grBOF; exit; end;
      FCurrentRecordF := WorkRec;
      Status := ReadARecord(Buffer);
      if status>0 then begin Result := grError; exit; end;

        with PRecInfo(Buffer + FRecInfoOfs)^ do
        begin
        BookmarkFlag := bfCurrent;
        RecordNumber := FCurrentRecordF;
        end;
        PInteger(Buffer + FBookmarkOfs)^ := FCurrentRecordF;

    If (Not FShowDelRec) and (buffer[0]='*') Then
      Begin { has record deleted}
      Accept := False;
      End
    Else
      Begin  {For filter event}
      Accept:=True;
      GetCalcFields(Buffer);
      FBufferForFilter := Buffer;
      If Assigned(OnFilterRecord) Then OnFilterRecord(self,Accept);
             {For Filter string}
      If Filtered Then  Accept := Accept and CheckFilter(FBufferForFilter);
              {For Master}
      If FOnMaster Then Accept := Accept and Links(FBufferForFilter);

      FBufferForFilter := nil;
      End;
 if (not Accept) and (GetMode = gmCurrent) then GetMode := gmNext;
 until Accept or (Result <> grOK);

    End Else Result := grError;
End;

Function TFoxDSet.GetRecordSize: Word;
    Begin
    Result:=FHeader.RecordLen;
    End;

    procedure TFoxDSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
    Begin
	messagedlg('TFoxDSet.InternalAddRecord', mtInformation, [mbYes],0);
    End;

procedure TFoxDSet.InternalInsert;
begin
      if GetBookmarkFlag(ActiveBuffer) = bfEOF then
              PInteger(ActiveBuffer + FBookmarkOfs)^ := FHeader.NumRecords+1;
end;

Procedure TFoxDSet.InternalFirst;
    Begin
    FCurrentRecordF := 0;
    End;

Procedure TFoxDSet.InternalLast;
    Begin
    FCurrentRecordF := FHeader.NumRecords+1;
      End;

Procedure TFoxDSet.InternalGotoBookmark(Bookmark: Pointer);
    Begin
    FCurrentRecordF := PInteger(Bookmark)^;
    End;

    procedure TFoxDSet.InternalHandleException;
    Begin
	messagedlg('TFoxDSet.InternalHandleException', mtInformation, [mbYes],0);
    End;

Procedure TFoxDSet.InternalInitFieldDefs;
    var Sname:string[31];
            I:integer;
    Begin
  if (FIsOpen) then begin
    FieldDefs.Clear;
    For I:=1 to FNFields do
          with FieldDefs.AddFieldDef do
    begin
      Sname:=trim(copy(FieldStruct[I].FieldName ,1,SizeOf(FieldStruct[I].FieldName)));
      FieldNo := I;
      Name := Sname;
      case FieldStruct[I].FieldType of
      'C': begin DataType := ftString; Size := FieldStruct[I].FieldLength; end;
      'Y': DataType := ftCurrency;
      'L': DataType := ftBoolean;
      'D': DataType := ftDate;
      'T': DataType := ftDateTime;
      'I': DataType := ftInteger;
      'G': DataType := ftInteger;
      'N','F': DataType := ftFloat;
      'B': if TableType = dBxMemo then DataType := ftMemo
                                  else DataType := ftFloat;
      'M': DataType := ftMemo;
      'P': DataType := ftMemo;
      end;
      Precision := FieldStruct[I].Decimals;
      Attributes := [faFixed];
      InternalCalcField := False;
      end;
                     end;
    End;

    procedure TFoxDSet.InternalInitRecord(Buffer: PChar);
    Begin
        FillChar(Buffer^,FRecordSize,' '); {only record}
    End;

Procedure TFoxDSet.InternalOpen;
    var openmode:integer;
        FullName:String;
    Begin
  if (FIsOpen) then Begin
                    DBFile.Free;
                    MemoFile.Free;
                    DBFile := nil;
                    MemoFile := nil;
                    FIsOpen:=False;
                    End;
  FullName := WorkDir + ExtractFileName(FTableName);
Try
  If FileExists(FullName + '.dbf') Then Begin

    if not FReadOnly then openmode:=fmOpenReadWrite or fmShareDenyNone
                     else openmode:=fmOpenRead or fmShareDenyNone;

    DBFile := TFileStream.Create(FullName + '.dbf', openmode);

{    if not ope(openmode) then Begin openmode:=fmOpenRead or fmShareDenyNone;
                                    ope(openmode);
                                    FReadOnly:=True;
                              End;}
    FillHeaderInfo;
    FIsOpen:=True;
    FCurrentRecordF:=0;
    FTypeDBF := FHeader.versionNumber;
    If (TableType = dBxMemo) and FileExists(FullName+'.dbt') Then Begin
      MemoFile := TFileStream.Create(FullName + '.dbt', openmode);
      MemoFile.Seek(0, soFromBeginning);
      MemoFile.ReadBuffer(FDBTHeader, Sizeof(FDBTHeader));
      FDBTHeader.BlockSize := 512;
    End;
    If ((FHeader.versionNumber = FoxMemo)or
        (FHeader.versionNumber = FoxV)) and FileExists(FullName+'.fpt') Then Begin
      MemoFile := TFileStream.Create(FullName + '.fpt', openmode);
      MemoFile.Seek(0, soFromBeginning);
      MemoFile.ReadBuffer(FDBTHeader, Sizeof(FDBTHeader));
          //back read
      invertDWORD(FDBTHeader.NextBlock);
      invertDWORD(FDBTHeader.BlockSize);
    End;
  FieldDefs.Updated := False;
  FieldDefs.Update;
  DefaultFieldClasses[ftMemo]:=TMemoFieldP;
  if DefaultFields then CreateFields;
  BindFields(True);
  InitBufferPointers(False);
  MasterFields := MasterFields;
  IndexFieldNames := IndexFieldNames;
  End else messagedlg('File not found '+ FullName, mtInformation, [mbYes],0);
Except
messagedlg('Error open '+ FullName, mtInformation, [mbYes],0);
End;
    End;

Procedure TFoxDSet.InternalPost;
    var buffer:PChar;
        source,tagert: integer;
Begin
  if State = dsEdit then WriteARecord(ActiveBuffer);
  if State = dsInsert then
   Begin
      Buffer := ActiveBuffer;
    tagert := PInteger(Buffer + FBookmarkOfs)^;
      inc(FHeader.NumRecords);
    source := FHeader.NumRecords;
      PInteger(Buffer + FBookmarkOfs)^ := FHeader.NumRecords;
    WriteARecord(Buffer);
      DbFile.Seek(0, soFromBeginning);
      DBFile.WriteBuffer(FHeader, Sizeof(FHeader));

      if EOF then
      FCurrentRecordF:=FHeader.NumRecords
      else
      if source <> tagert then MoveRecs(source, tagert);

   End;
End;

Procedure TFoxDSet.InternalClose;
    Begin
    if DefaultFields then DestroyFields;
    DBFile.Free;
    MemoFile.Free;
    DBFile := nil;
    MemoFile := nil;
    FIsOpen:=False;
    End;

procedure TFoxDSet.InternalDelete;
Var d,b:PChar;
    Begin
    d:='*';
    b:=ActiveBuffer;
    UpdateFieldData(0,b, d);
    WriteARecord(ActiveBuffer);
    inc(FCurrentRecordF);
End;

Procedure TFoxDSet.InternalSetToRecord(Buffer: PChar);
    Begin
    InternalGotoBookmark(Buffer + FBookmarkOfs);
    End;

Function TFoxDSet.IsCursorOpen: Boolean;
    Begin
    Result:=FIsOpen;
    End;

    procedure TFoxDSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  RecBuf: PChar;
begin
  with Field do
  begin
//    if not (State in dsWriteModes) then DatabaseError(SNotEditing, Self);
//    if (State = dsSetKey) and ((FieldNo < 0) or (FIndexFieldCount > 0) and
//      not IsIndexField) then DatabaseErrorFmt(SNotIndexField, [DisplayName]);
    GetActiveRecBuf(RecBuf);
    if FieldNo > 0 then
    begin
      if State = dsCalcFields then DatabaseError(SNotEditing, Self);
      if ReadOnly and not (State in [dsSetKey, dsFilter]) then
        DatabaseErrorFmt(SFieldReadOnly, [DisplayName]);
      Validate(Buffer);
      if FieldKind <> fkInternalCalc then
      begin
        UpdateFieldData(FieldNo, RecBuf, Buffer);
      end;
    end else {fkCalculated, fkLookup}
    begin
      Inc(RecBuf, FRecordSize + Offset);
      Boolean(RecBuf[0]) := LongBool(Buffer);
      if Boolean(RecBuf[0]) then Move(Buffer^, RecBuf[1], DataSize);
    end;
    if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
      DataEvent(deFieldChange, Longint(Field));
  end;
end;

Procedure TFoxDSet.SetTableName(Name: String);
    Begin
  FTableName:= ChangeFileExt(Name,'');
//  If FileExists(FTableName + '.dbf') Then
  FWorkDir:=ExtractFileDir(ExpandFileName(FTableName))+'\';

    End;

function TFoxDSet.ope(i:integer):boolean;
begin
    result:=false;
  try
    DBFile := TFileStream.Create(FTableName + '.dbf', i);
    result:=true;
  except
    on EFOpenError do result:=false;
  else
  end;
end;

procedure TFoxDSet.SetReadOnly(Value: Boolean);
begin
  FReadOnly := Value;
end;

Procedure TFoxDSet.FillHeaderInfo;
Var
  TmpC, TmpA: Integer;
Begin
  {Get Header Info}
  Try
    DBFile.Seek(0, soFromBeginning);
    DBFile.ReadBuffer(FHeader, Sizeof(FHeader));
    FNFields:=Trunc((FHeader.HeaderLen - 32) / 32);
    {GetFieldInfo}
    TmpA := 2;
    For TmpC := 1 To FNFields Do Begin
      DBFile.Seek(TmpC * 32, soFromBeginning);
      DBFile.ReadBuffer(FieldStruct[TmpC], Sizeof(FieldStruct[TmpC]));
      If FieldStruct[TmpC].FieldName[1] = #13 then begin
                                                FNFields := TmpC-1;
                                                break;
                                                end;
      FieldStruct[TmpC].Address := TmpA;
      TmpA := TmpA + FieldStruct[TmpC].FieldLength;
    End;
 //     FieldStruct[0].FieldName:='Del         ';
      FieldStruct[0].FieldType:='C';
      FieldStruct[0].Address := 0;
      FieldStruct[0].FieldLength:=1;
      FieldStruct[0].Decimals:=0;
  Except
  End;
End;

Function TFoxDSet.ReadARecord(RecordBuffer:PChar):integer;
Begin
  try
    Result:=1;
    if (FCurrentRecordF>0) and (FCurrentRecordF<=FHeader.NumRecords) then
    begin
    DBFile.Seek((FHeader.HeaderLen) + ((FCurrentRecordF - 1) * FHeader.RecordLen), soFromBeginning);
    DBFile.ReadBuffer(RecordBuffer^, FHeader.RecordLen);
    if assigned(FOnEncode) then FOnEncode(RecordBuffer,FHeader.RecordLen,bmRead);
    end;
    Result:=0;
  finally
  end;
End;

Procedure TFoxDSet.WriteARecord(RecordBuffer:PChar);
var NRecord:integer;
Begin
  Try
    NRecord := PInteger(RecordBuffer + FBookmarkOfs)^;
    DBFile.Seek((FHeader.HeaderLen) + ((NRecord - 1) * FHeader.RecordLen), soFromBeginning);
    if assigned(FOnEncode) then FOnEncode(RecordBuffer,FHeader.RecordLen,bmWrite);
    DBFile.WriteBuffer(RecordBuffer^, FHeader.RecordLen);
  Except
  End;
End;

function TFoxDSet.GetFieldData(FieldNo: Integer; Buffer: Pointer): Boolean;
var
  RecBuf: PChar;
  InBuf: PChar;
  F:PDouble;
  I:Pint64;
  Ismoll:Pinteger;
  TS: TTimeStamp;
  B: ^WordBool;
begin
    if buffer = nil then begin Result := False; exit; end;
    if FBufferForFilter <> nil then
      begin RecBuf := FBufferForFilter; Result := True; end
    else
    Result := GetActiveRecBuf(RecBuf);
    if Result then
    begin
      Inc(RecBuf, FieldStruct[FieldNo].Address-1);
    case FieldStruct[FieldNo].FieldType of
    'N','F': begin                {Length digital in DBF great Buffer}
          InBuf := Allocmem(FieldStruct[FieldNo].FieldLength+1);
         try
          Move(RecBuf[0], InBuf^, FieldStruct[FieldNo].FieldLength);
          InBuf[FieldStruct[FieldNo].FieldLength]:=#0;
          try
          If (FieldStruct[FieldNo].FieldLength<>0) and
              (length(trim(InBuf))>0)  then Double(Buffer^) := strtofloat(InBuf)
          else Double(Buffer^) := 0;
          except
          on EConvertError do Double(Buffer^) := 0;
          end;
         finally
         FreeMem(InBuf);
         end;
         end;
      'C': begin
           Move(RecBuf[0], Buffer^, FieldStruct[FieldNo].FieldLength);
           PChar(Buffer)[FieldStruct[FieldNo].FieldLength]:=#0;
           end;
      'Y': begin
           Move(RecBuf[0], Buffer^, FieldStruct[FieldNo].FieldLength);
           I := Buffer;
           F := Buffer;
           F^ := I^/10000;
           end;
      'T': begin
           Move(RecBuf[0], Buffer^, FieldStruct[FieldNo].FieldLength);
           Ismoll := Buffer;
           TS.Date := Ismoll^-1721425;
           inc(Ismoll,1);
           TS.Time := Ismoll^;
           F := Buffer;
           F^ := TimeStampToDateTime(TS);
//           F^ := Ismoll^-2415019;
           end;
      'L': begin
           B := Buffer;
           If RecBuf[0]='T' then B^ := True;
           If RecBuf[0]='F' then B^ := False;
           end;
    else
        Move(RecBuf[0], Buffer^, FieldStruct[FieldNo].FieldLength);
    end;
    end;
end;

function TFoxDSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  RecBuf: PChar;
begin
  if Field.FieldNo >= 0 then
    Result := GetFieldData(Field.FieldNo, Buffer)
  else
  begin
        Result := GetActiveRecBuf(RecBuf);
    if Result and (State in [dsBrowse, dsEdit, dsInsert, dsCalcFields]) then
    begin
      Inc(RecBuf, FRecordSize + Field.Offset);
      Result := Boolean(RecBuf[0]);
      if Result and (Buffer <> nil) then
        Move(RecBuf[1], Buffer^, Field.DataSize);
    end;
  end;
end;
function TFoxDSet.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
  case State of
    dsBrowse: if IsEmpty then RecBuf := nil else RecBuf := ActiveBuffer;
    dsEdit, dsInsert: RecBuf := ActiveBuffer;
    dsCalcFields: RecBuf := CalcBuffer;
  else
    RecBuf := nil;
  end;
  Result := RecBuf <> nil;
end;

procedure TFoxDSet.InitBufferPointers(GetProps: Boolean);
begin
  if FIsOpen then
  begin
    BookmarkSize := 10;
    FRecordSize := FHeader.RecordLen;
  end;
  FBlobCacheOfs := FRecordSize + CalcFieldsSize;
  FRecInfoOfs := FBlobCacheOfs + BlobFieldCount * SizeOf(Pointer);
  FBookmarkOfs := FRecInfoOfs + SizeOf(TRecInfo);
  FRecBufSize := FBookmarkOfs + BookmarkSize;
end;

{blob}

function TFoxDSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
Result := TMemoStream.Create(Field as TBlobField, Mode);
end;

     procedure TFoxDSet.DataConvert(Field: TField; Source, Dest: Pointer; ToNative: Boolean);
begin
try
  case Field.DataType of
    ftDate:
      if ToNative then
{write} Move(PChar(FormatDateTime('yyyymmdd',TDateTime(Source^)))[0],Dest^,8) else
        begin
{read}
        TDateTime(Dest^) := EncodeDate(StrToIntDef(copy(PChar(Source),1,4),1),
                                       StrToIntDef(copy(PChar(Source),5,2),1),
                                       StrToIntDef(copy(PChar(Source),7,2),1));
        end;
    else move(Source^,Dest^,8);
 end;
except
On EConvertError do TDateTime(Dest^) := 1;
end;
end;
    procedure TFoxDSet.SetShowDelRec(Value: Boolean);
begin
FShowDelRec := Value;
If FIsOpen then Refresh;
end;

 {filter}
function slipword(var s:string):string;
var i:integer;
begin
s:=trim(s);
try
if Length(s)>0 then begin
        if s[1]='''' then begin i:=pos('''',PChar(@s[2]));
                          inc(i,2);
                          end
                     else
                     i:=pos(' ',s);
        if i=0 then begin
                 result := s;
                 s:='';
                 end
                else begin
                 result:=copy(s,1,i);
                 s:=copy(s,i,length(s));
                 end;
end
else result := s;
except
result := s;
s:='';
end;
end;
function TFoxDSet.AlignString( var i:integer; F:boolean):boolean;
var    Fi,Fi1:TField;
       Elem,Elem1,Elem2 : ^Variant;
begin
Fi:=nil;
Fi1:=nil;
result:=True;
Elem := @FFilterData[i].val;
Elem1 := @FFilterData[i+1].val;
Elem2 := @FFilterData[i+2].val;
case String(Elem1^)[1] of
'>',
'<',
'=': begin
      {If >= <= <>}
      If Length(String(Elem1^))>1 then
        begin
        If (String(Elem1^)[1]='<') and (String(Elem1^)[2]='=') then Elem1^:='L';
        If (String(Elem1^)[1]='>') and (String(Elem1^)[2]='=') then Elem1^:='G';
        If (String(Elem1^)[1]='>') and (String(Elem1^)[2]='<') then Elem1^:='N';
        If (String(Elem1^)[1]='<') and (String(Elem1^)[2]='>') then Elem1^:='N';
        end;
     {Def first operand}      {Field}
        if String(Elem^)[1]='''' then begin
        if not (foCaseInsensitive in FilterOptions) then Elem^:=AnsiUpperCase(String(Elem^));
        FFilterData[i].IsPole := False;
        Elem^ := copy(Elem^,2,Length(Elem^)-2);
        if not (foNoPartialCompare in FilterOptions) then
        If (String(Elem^)[1]='*')and(String(Elem^)[Length(Elem^)]='*') then
                begin Elem1^:='1';
                Elem^ := copy(Elem^ ,2,Length(Elem^)-2);
                end
        else If (String(Elem^)[1]='*') then
                begin Elem1^:='2';
                Elem^ := copy(Elem^ ,2,Length(Elem^)-1);
                end
             else If (String(Elem^)[Length(Elem^)]='*') then
                begin Elem1^:='3';
                Elem^ := copy(Elem^ ,1,Length(Elem^)-1);
                end;
        FFilterData[i].count:=Length(Elem^);
          If assigned(FOnConvert) then FOnConvert(@(string(Elem^)[1]),@(string(Elem^)[1]), True)
          else CharToOemBuff(@(string(Elem^)[1]),@(string(Elem^)[1]),FFilterData[i].count);
           {case FCodePagePost of
           OEM: CharToOemBuff(@(string(Elem^)[1]),@(string(Elem^)[1]),FFilterData[i].count);
           UserConvert: if Assigned(FOnConvert) then
                        FOnConvert(@(string(Elem^)[1]),FFilterData[i].count,bmWrite);
           end;}
        end
        else begin            {Const}
        Fi:=Fields.FindField(Elem^);
        if Fi<>nil then begin
        FFilterData[i].IsPole := True;
        FFilterData[i].offset:=FieldStruct[Fi.FieldNo].Address;
        FFilterData[i].count:=Fi.FieldNo;
        end else messagedlg('Fied '+Elem^+' not found ', mtInformation, [mbYes],0);
        end;

     {Def second operand}      {Field}
        if string(Elem2^)[1]='''' then begin
        if not (foCaseInsensitive in FilterOptions) then Elem2^:=AnsiUpperCase(String(Elem2^));
        FFilterData[i+2].IsPole := False;
        Elem2^ := copy(Elem2^ ,2,Length(Elem2^)-2);
        if not (foNoPartialCompare in FilterOptions) then
        If (String(Elem2^)[1]='*')and(String(Elem2^)[Length(Elem2^)]='*') then
                begin Elem1^:='5';
                Elem2^ := copy(Elem2^ ,2,Length(Elem2^)-2);
                end
        else If (String(Elem2^)[1]='*') then
                begin Elem1^:='6';
                Elem2^ := copy(Elem2^ ,2,Length(Elem2^)-1);
                end
             else If (String(Elem2^)[Length(Elem2^)]='*') then
                begin Elem1^:='7';
                Elem2^ := copy(Elem2^ ,1,Length(Elem2^)-1);
                end;
        FFilterData[i+2].count:=Length(Elem2^);
          If assigned(FOnConvert) then FOnConvert(@(string(Elem2^)[1]),@(string(Elem2^)[1]), True)
          else CharToOemBuff(@(string(Elem2^)[1]),@(string(Elem2^)[1]),FFilterData[i+2].count);
           {case FCodePagePost of
           OEM: CharToOemBuff(@(string(Elem2^)[1]),@(string(Elem2^)[1]),FFilterData[i+2].count);
           UserConvert: if Assigned(FOnConvert) then
                        FOnConvert(@(string(Elem2^)[1]),FFilterData[i+2].count,bmWrite);
           end;}
        end
        else begin
        Fi1:=Fields.FindField(Elem2^);
        if Fi1<>nil then begin
        FFilterData[i+2].IsPole := True;
        FFilterData[i+2].offset:=FieldStruct[Fi1.FieldNo].Address;
        FFilterData[i+2].count:=Fi1.FieldNo;
        end else messagedlg('Fied '+Elem2^+' not found ', mtInformation, [mbYes],0);
        end;
         if FFilterData[i].IsPole and (not FFilterData[i+2].IsPole) then
           case FieldStruct[Fi.FieldNo].FieldType of
           'N','F': FFilterData[i+2].val := strtofloat(FFilterData[i+2].val);
           'D': FFilterData[i+2].val := FormatDateTime('yyyymmdd', strtodate(FFilterData[i+2].val));
           end;
         if FFilterData[i+2].IsPole and (not FFilterData[i].IsPole) then
           case FieldStruct[Fi1.FieldNo].FieldType of
           'N','F': FFilterData[i].val := strtofloat(FFilterData[i].val);
           'D': FFilterData[i].val := FormatDateTime('yyyymmdd', strtodate(FFilterData[i].val));
           end;
        inc(i,2) end;
'o': begin inc(i,2); result:=AlignString(i,True) or F;    end;
'a': begin inc(i,2); result:=AlignString(i,True) and F;   end;
end;
i:=i;
if FFilterDataCount>(i+1) then result:=AlignString(i,Result);
i:=i;
end;

function TFoxDSet.GreatLess( var i:integer; F:boolean):boolean;
begin
case string(FFilterData[i+1].val)[1] of
'1': begin result:= pos(FFilterData[i].val,FFilterData[i+2].val)>0; inc(i,2) end;
'2': begin result:= (Length(trim(FFilterData[i+2].val))-Length(trim(FFilterData[i].val))+1-pos(FFilterData[i].val,FFilterData[i+2].val))=0; inc(i,2) end;
'3': begin result:= pos(FFilterData[i].val,FFilterData[i+2].val)=1; inc(i,2) end;
'5': begin result:= pos(FFilterData[i+2].val,FFilterData[i].val)>0; inc(i,2) end;
'6': begin result:= (Length(trim(FFilterData[i].val))-Length(trim(FFilterData[i+2].val))+1-pos(FFilterData[i+2].val,FFilterData[i].val))=0; inc(i,2) end;
'7': begin result:= pos(FFilterData[i+2].val,FFilterData[i].val)=1; inc(i,2) end;
'>': begin result:=FFilterData[i].val>FFilterData[i+2].val; inc(i,2) end;
'<': begin result:=FFilterData[i].val<FFilterData[i+2].val; inc(i,2) end;
'=': begin result:=FFilterData[i].val=FFilterData[i+2].val; inc(i,2) end;
'G': begin result:=FFilterData[i].val>=FFilterData[i+2].val; inc(i,2) end;
'L': begin result:=FFilterData[i].val<=FFilterData[i+2].val; inc(i,2) end;
'N': begin result:=FFilterData[i].val<>FFilterData[i+2].val; inc(i,2) end;
'o': begin inc(i,2); result:=GreatLess(i,True) or F;    end;
'a': begin inc(i,2); result:=GreatLess(i,True) and F;   end;
else result:=True;
end;
if FFilterDataCount>(i+1) then result:=GreatLess(i,Result);
end;

function TFoxDSet.CheckFilter(buffer:PChar):boolean;
var i,j:integer;
    Tmp: PChar;
  Ismoll:Pinteger;
  TS: TTimeStamp;
begin
For i:=0 to FFilterDataCount-1 do
begin
if FFilterData[i].IsPole then
        begin
        Tmp:=@buffer[FFilterData[i].offset-1];
        case FieldStruct[FFilterData[i].count].FieldType of
   'N','F': begin
            FFilterData[i].val := copy(Tmp,0,FieldStruct[FFilterData[i].count].FieldLength);
            FFilterData[i].val := strtofloat(FFilterData[i].val);
            end;
   'I': FFilterData[i].val := PInteger(Tmp)^;
   'B': if TableType = Fox30 then FFilterData[i].val := PDouble(Tmp)^;
   'Y': FFilterData[i].val := PCurrency(Tmp)^;
   'T':    begin
           Ismoll := PInteger(Buffer);
           TS.Date := Ismoll^-1721425;
           inc(Ismoll,1);
           TS.Time := Ismoll^;
           FFilterData[i].val := TimeStampToDateTime(TS);
           end;
        else if not (foCaseInsensitive in FilterOptions) then
                FFilterData[i].val := trim(AnsiUpperCase(copy(Tmp,0,FieldStruct[FFilterData[i].count].FieldLength)))
             else
                FFilterData[i].val := trim(copy(Tmp,0,FieldStruct[FFilterData[i].count].FieldLength));
        end;
{         if FFilterData[i].IsPole and (not FFilterData[i+2].IsPole) then
           case FieldStruct[FFilterData[i].count].FieldType of
           'N','F': FFilterData[i].val := strtofloat(FFilterData[i].val);
//           'D': FFilterData[i].val := FormatDateTime('yyyymmdd', strtodate(FFilterData[i].val));
           end;}
        end;
end;
i:=0;
result:= GreatLess(i,True);
end;

procedure TFoxDSet.SetFilterText(const Value: string);
var i:integer;
    s:string;
begin
    CheckBrowseMode;
   inherited SetFilterText(Value);
  s:= Value;
  i:=0;
  while length(slipword(s))>0 do inc(i);
  SetLength(FFilterData,i);
  FFilterDataCount:=i;
  s:= Value;
  i:=0;
  while length(s)>0 do begin
        FFilterData[i].val:=trim(slipword(s));
        FFilterData[i].count:=length(FFilterData[i].val);
        inc(i);
        end;
{For speed}
i:=0;
AlignString(i,True);
end;

procedure TFoxDSet.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if Filtered <> Value then
    begin
      inherited SetFiltered(Value);
    end;
    First;
  end else
    inherited SetFiltered(Value);
end;

{for user programming}
function TFoxDSet.GetRecordCount: Longint;
begin
  if FIsOpen then Result := FHeader.NumRecords
             else Result := -1;
end;

function TFoxDSet.GetRecNo: Integer;
var
  BufPtr: PChar;
begin
  if State = dsCalcFields then
    BufPtr := CalcBuffer else
    BufPtr := ActiveBuffer;
  Result := PRecInfo(BufPtr + FRecInfoOfs).RecordNumber;
end;
   {on ShowDelRec, OnFilter}
procedure TFoxDSet.SetRecNo(Value: Integer);
begin
  if (Value <> RecNo) then
  begin
    CheckBrowseMode;
    DoBeforeScroll;
    FCurrentRecordF:=Value;
    Resync([rmCenter]);
    DoAfterScroll;
  end;
end;

{TMemoStream}

constructor TMemoStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
var
  TmpS:String;
  Offset:Pchar;
  Buff: Array[1..512] Of Char;
  Flen,i,TmpL,b:integer;
  FileOffset: PInteger;
begin
    FDataSet := Field.DataSet as TFoxDSet;
    FField := Field;
    FSize := 0;
    FMode := Mode;
    if FDataSet.MemoFile = nil then begin messagedlg('Memo file not open', mtInformation, [mbYes],0); exit; end;
{READ}
  if Mode = bmRead then
    begin
    if FField is TMemoFieldP then TMemoFieldP(FField).FieldKnow := False;
    Field.Modified:=False;
    Flen:=FDataSet.FieldStruct[Field.FieldNo].FieldLength;
    SetLength(TmpS,Flen);
      Offset:=FDataSet.ActiveBuffer;
      inc(Offset, FDataSet.FieldStruct[Field.FieldNo].Address-1);
      i:=0;
      FileOffset := @i;
      if Flen = 4 then FileOffset := PInteger(Offset);
      if Flen = 10 then
          FileOffset^ := StrToIntDef(copy(Offset,0, Flen), 0);
    with FDataSet do if FileOffset^=0 then begin
                                        FSize:=0;
                                        StartPos := MemoFile.Position
                                       end
    else begin
    MemoFile.Seek(FileOffset^ * integer(FDBTHeader.BlockSize), soFromBeginning);
    MemoFile.Read (buff,4);
    MemoFile.Read (nlen,4);
    If (TableType = dBxMemo) Then Begin
                                  MemoFile.Read (buff,4);
                                  MemoFile.Read (buff,4);
                                  End
                             Else Begin
                                  invertDWORD(nlen);
                                  End;
    FSize:=nlen;
    StartPos := MemoFile.Position;
     {type graphics}
     if (FSize>1) then begin
       MemoFile.Read(TMemoFieldP(FField).TypeGraph,2);
       MemoFile.Position := StartPos;

       case TMemoFieldP(FField).TypeGraph of
       $4D42{bmp},
       $D8FF{jpg},
       $CDD7{wmf},
       $0001{emf},
       $0000:{ico} begin TMemoFieldP(FField).FieldKnow := True;
                   end
       else begin TMemoFieldP(FField).FieldKnow := False;
            end;
       end;
                        end;
    end;
    end;
{WRITE}
if (Mode = bmWrite) then
    if (FDataSet.Tabletype = dBxMemo) then messagedlg('Not write Memo this TableType ', mtInformation, [mbYes],0)
    else
    begin
    Field.Modified:=True;
    with FDataSet do begin
       {align memo}
       MemoFile.Seek( 0, soFromEnd);
       if ( DWord(MemoFile.Size) mod FDBTHeader.BlockSize )>0 then begin
                  {TmpL - count byte to end blocksize}
       TmpL := FDBTHeader.BlockSize - ( DWord(MemoFile.Size) mod FDBTHeader.BlockSize );
       b := 0;   For i := 1 to TmpL do MemoFile.Write ( b, 1 );
                                                                   end;
    MemoFile.Write (i, 4);MemoFile.Write (i, 4);{write hand record}
    StartPos := MemoFile.Position;
    end;
//    if not (FDataSet.State in [dsCalcFields, dsFilter, dsNewValue]) then
//      FDataSet.DataEvent(deFieldChange, Longint(Field));
    end;
end;

function TMemoStream.Read(var Buffer; Count: Longint): Longint;
begin
 TFoxDSet(FDataSet).MemoFile.Read(Buffer,Count);
if TMemoFieldP(FField).Transliterate then TFoxDSet(FDataSet).Translate(PChar(@Buffer),PChar(@Buffer),False);
Result:= Count;
end;

function TMemoStream.Write(const Buffer; Count: Longint): Longint;
begin
if TMemoFieldP(FField).Transliterate then TFoxDSet(FDataSet).Translate(PChar(@Buffer),PChar(@Buffer),True);
if (FDataSet.TableType <> dBxMemo) then
TFoxDSet(FDataSet).MemoFile.Write(Buffer,Count);
Result:=Count;
end;

function TMemoStream.Seek(Offset: Longint; Origin: Word): Longint;
 begin
  case Origin of
    0: TFoxDSet(FDataSet).MemoFile.Seek(StartPos + Offset,0);
    1: TFoxDSet(FDataSet).MemoFile.Seek(Offset,1);
    2: TFoxDSet(FDataSet).MemoFile.Seek(StartPos+FSize-Offset,0);
  end;
  Result := TFoxDSet(FDataSet).MemoFile.Position - StartPos;
 end;
destructor TMemoStream.Destroy;
 var i: integer;
     Block: string;
     RecordBuffer:PChar;
 begin
if (FMode = bmWrite)and(FDataSet.TableType <> dBxMemo) then
    with FDataSet do begin
         {write head record memo}
    MemoFile.Seek(StartPos-8, soFromBeginning);
    i := 1;                      invertDWORD(DWORD(i));  MemoFile.Write (i, 4);
    i := MemoFile.Size-StartPos; invertDWORD(DWORD(i));  MemoFile.Write (i, 4);
         {write head file memo}
    MemoFile.Seek(0, soFromEnd);
    FDBTHeader.NextBlock := DWord(MemoFile.Position) div FDBTHeader.BlockSize;
    if (DWord(MemoFile.Position) mod FDBTHeader.BlockSize) > 0 then
                                            inc( FDBTHeader.NextBlock );
    invertDWORD(DWORD(FDBTHeader.BlockSize));
    invertDWORD(DWORD(FDBTHeader.NextBlock));
    MemoFile.Seek(0, soFromBeginning);
    MemoFile.WriteBuffer(FDBTHeader, sizeof(TDBTHeader));
    invertDWORD(DWORD(FDBTHeader.BlockSize));
    invertDWORD(DWORD(FDBTHeader.NextBlock));
          {write field in record DBF}
    Block := IntToStr( DWORD(StartPos) div FDBTHeader.BlockSize );
     If (FHeader.versionNumber = FoxMemo) Then
         Block := StringOfChar(' ',10-Length(Block))+Block;
     If (FHeader.versionNumber = FoxV)    Then
         PInteger(Block)^ := DWORD(StartPos) div FDBTHeader.BlockSize;
    GetActiveRecBuf(RecordBuffer);
    UpdateFieldData(FField.FieldNo,RecordBuffer, @Block[1]);
          if state <> dsInsert Then SetState(dsEdit);
          SetModified(True);
//    WriteARecord(RecordBuffer);
      //end write DBF
    end;
 inherited;
 end;

{Master}
procedure TFoxDSet.MasterChanged(Sender: TObject);
begin
  CheckBrowseMode;
  FOnMaster := True;
    First;
end;
procedure TFoxDSet.MasterDisabled(Sender: TObject);
begin
    CheckBrowseMode;
  FOnMaster := False;
    First;
end;
function TFoxDSet.Links(Buffer: PChar): boolean;
var b,b1:PChar;
begin
        b:= TFoxDSet(FMasterLink.DataSource.DataSet).ActiveBuffer;
        inc(b,FMasterOffcet);
        b1:= Buffer;
        inc(b1,FIndexOffcet);
        if StrLComp(b,b1,FMasterLen)=0 then Result := True
                                       else Result := False;
end;

 function TFoxDSet.GetDataSource: TDataSource;
begin
  Result := FMasterLink.DataSource;
end;

procedure TFoxDSet.SetDataSource(Value: TDataSource);
begin
  if IsLinkedTo(Value) then DatabaseError(SCircularDataLink, Self);
  FMasterLink.DataSource := Value;
end;
function TFoxDSet.GetMasterFields: string;
begin
  Result := FMasterLink.FieldNames;
end;
procedure TFoxDSet.SetMasterFields(const Value: string);
var i:integer;
begin
if FMasterLink.DataSource<> nil then
begin
        for i:=0 to TFoxDSet(FMasterLink.DataSource.DataSet).FieldDefs.count-1 do
        begin
        IF TFoxDSet(FMasterLink.DataSource.DataSet).FieldDefs[i].Name = Value then
        begin
        FMasterLink.Fields.Add((FMasterLink.DataSource.DataSet).FieldDefs[i]);
        FMasterLink.FieldNames := Value;
        FMasterOffcet := TFoxDSet(FMasterLink.DataSource.DataSet).FieldStruct[i+1].Address-1;
        FMasterLen := TFoxDSet(FMasterLink.DataSource.DataSet).FieldStruct[i+1].FieldLength;
        end
        end;
end;
  FMasterLink.FieldNames := Value;
end;
function TFoxDSet.GetIndexFieldNames: string;
begin
Result := FIndexName;
//  if FFieldsIndex then Result := FIndexName else Result := '';
end;
procedure TFoxDSet.SetIndexFieldNames(const Value: string);
var i: integer;
begin
        for i:=0 to FieldDefs.count-1 do
        begin
        IF FieldDefs[i].Name = Value then
        begin
        FIndexOffcet := FieldStruct[i+1].Address-1;
        FIndexLen := FieldStruct[i+1].FieldLength;
        FIndexName := Value;
        end
        end;
//  SetIndex(Value, Value <> '');
end;
procedure TFoxDSet.CreateTable;
Var     HeaderNew: THeader;
        MHeaderNew: TDBTHeader;
        MExist: boolean;
        FieldS: TFieldStruct;
        RF: Byte;
        FileTmp: TFileStream;
        i,j,Sum: integer;
begin
If not FIsOpen then begin
{Write DBF Head}
FillChar(HeaderNew,SizeOf(HeaderNew),0);
with HeaderNew do begin
    if FTypeDBF<>0 then VersionNumber := FTypeDBF
                   else VersionNumber := $F5;
    LastUpdated[1] := StrToInt(FormatDateTime('yy',Date()));
    LastUpdated[2] := StrToInt(FormatDateTime('mm',Date()));
    LastUpdated[3] := StrToInt(FormatDateTime('dd',Date()));
    NumRecords := 0;
    end;
FileTmp := TFileStream.Create(FTableName + '.dbf', fmCreate);
FileTmp.Seek(0, soFromBeginning);
FileTmp.WriteBuffer(HeaderNew,SizeOf(HeaderNew));
    Sum:=1; j:=0;
    for i:=0 to FieldDefs.count-1 do
    if GetFieldS(FieldDefs[i], FieldS) then begin
    begin FieldS.Address := Sum;
    Sum := Sum + FieldS.FieldLength; inc(j);
    FileTmp.WriteBuffer(FieldS,SizeOf(FieldS)); end;
    end;
RF:=13; FileTmp.WriteBuffer(RF,1);
HeaderNew.RecordLen := Sum;
HeaderNew.HeaderLen := SizeOf(HeaderNew) + j * SizeOf(FieldS)+1;
    MExist:=False;
    for i:=0 to FieldDefs.count-1 do
    if FieldDefs[i].DataType = ftMemo then MExist:=true;
    If MExist and (FTypeDBF = $30) then HeaderNew.MdxFlag := 2
                                   else HeaderNew.MdxFlag := 0;
RF:=0;  {the relative path 0}
if TableType = Fox30 then begin
                for i:=1 to 263 do FileTmp.WriteBuffer(RF,1);
                HeaderNew.HeaderLen := HeaderNew.HeaderLen +263;
                end;
FileTmp.Seek(0, soFromBeginning);
FileTmp.WriteBuffer(HeaderNew,SizeOf(HeaderNew));
FileTmp.Free;
{Memo Create}
If MExist then begin
        FillChar(MHeaderNew,SizeOf(MHeaderNew),0);
        MHeaderNew.NextBlock := 8;
        MHeaderNew.BlockSize := 64;
        invertDWORD(MHeaderNew.NextBlock);
        invertDWORD(MHeaderNew.BlockSize);
        FileTmp := TFileStream.Create(FTableName + '.fpt', fmCreate);
        FileTmp.WriteBuffer(MHeaderNew,SizeOf(MHeaderNew));
        FileTmp.Free;
        end
        else if FileExists(FTableName + '.fpt') then DeleteFile(FTableName + '.fpt');


end;
end;

    function TFoxDSet.GetTableCreate: Boolean;
    begin
    Result := FTableCreate;
    end;
    procedure TFoxDSet.SetTableCreate( Value: Boolean);
    var s:string;
        i: integer;
    begin
    If (not Active) and (not FIsOpen) and (not FTableCreate) and (Length(TableName)>0) and Value then
        begin
        s:='Creating Table '''+ FTableName + '''. Fields:';
        for i:=0 to FieldDefs.Count-1 do
              s:= s+FieldDefs[i].Name+',';
        if (messagedlg(s+'.?', mtConfirmation, [mbYes,mbNo],0)=mrYes) then
        begin
           if FileExists(FTableName + '.dbf') and
           (messagedlg('The table exist, destroy ?', mtConfirmation, [mbYes,mbNo],0)=mrYes) then
           CreateTable
           else CreateTable;
        end;
        end;
//    else
//        FTableCreate := Value;
    end;

procedure TFoxDSet.CopyToFile( Name: string);
var DataS: TFoxDSet;
    TmpName: string;
begin
  TmpName:= ChangeFileExt(Name,'');
DataS := TFoxDSet.Create(self);
Try
DataS.TableName := TmpName;                      {Create}
           if not FileExists(TmpName + '.dbf') then
                begin
                DataS.FieldDefs := FieldDefs;
                DataS.FTypeDBF := FTypeDBF;
                DataS.CreateTable;
                end;
CopyTo(DataS);
Finally
DataS.Destroy;
end;
end;

procedure TFoxDSet.CopyTo( Dest: TFoxDSet);
var    I,J: Integer;
begin
Dest.Active := True;
Dest.FTranlateOn := False;
FTranlateOn := False;
First;
while not Eof do
 begin
 Dest.Append;
 For I:=0 to Fields.Count-1 do
 For J:=0 to Dest.Fields.Count-1 do
 If FieldDefs[I].Name = Dest.FieldDefs[J].Name then
        begin
        Dest.Fields[J].Value := Fields[J].Value;
        end;
 Dest.Post;
 Next;
 end;
Dest.FTranlateOn := True;
FTranlateOn := True;
end;

procedure TFoxDSet.AppendFromFile( Name: string);
var DataS: TFoxDSet;
    TmpName: string;
begin
  TmpName:= ChangeFileExt(Name,'');
if FileExists(TmpName + '.dbf') then begin
DataS := TFoxDSet.Create(self);
Try
DataS.TableName := TmpName;
AppendFrom(DataS);
Finally
DataS.Destroy;
end;
End
else messagedlg('File '+Name+' not found', mtInformation, [mbYes],0);
end;

procedure TFoxDSet.AppendFrom( Source: TFoxDSet);
var I,J: Integer;
begin
Source.Active := True;
Source.First;
FTranlateOn := False;
Source.FTranlateOn := False;
while not Source.Eof do
 begin
 Append;
 For I:=0 to Source.Fields.Count-1 do
 For J:=0 to Fields.Count-1 do
 If FieldDefs[I].Name = Source.FieldDefs[J].Name then
        begin
        Fields[J].Value := Source.Fields[J].Value;
        end;
 Post;
 Source.Next;
 end;
Source.FTranlateOn := True;
FTranlateOn := True;
end;

procedure TFoxDSet.Pack;
var NameFile,TmpName: string;
    J: integer;
    TypeMemo: string;
    Arh,res: boolean;
begin
Arh := Active;
if not Arh then Active;
if (Tabletype = dBxMemo) then TypeMemo := '.dbt'
                         else TypeMemo := '.fpt';
NameFile := WorkDir + ExtractFileName(FTableName);

j:=1; {copy to tmp}
while FileExists(WorkDir + 'tmp'+IntToStr(J)+'.dbf') or
      FileExists(WorkDir + 'tmp'+IntToStr(J)+TypeMemo) do Inc(J);
      TmpName := WorkDir + 'tmp'+IntToStr(J);  {If Run-time changed directory}
 Filtered:=False;
 CopyToFile(TmpName);
 Active := False;

 {delete old buckup}
 if FileExists(NameFile+'.bak') then DeleteFile(NameFile+'.bak');
 if FileExists(NameFile+'.tbk') then DeleteFile(NameFile+'.tbk');
 {Create Buckup}
 Active := False;
 res:=RenameFile(NameFile+'.dbf',NameFile+'.bak');
 if FileExists(NameFile+TypeMemo) then
       res:=res and RenameFile(NameFile+TypeMemo,NameFile+'.tbk');

if not res then raise Exception.Create('Unable to create backup file.');

 {tmp to dbf}
 RenameFile(TmpName+'.dbf',NameFile+'.dbf');
 if FileExists(TmpName+TypeMemo) then
        RenameFile(TmpName+TypeMemo,NameFile+TypeMemo);
Active := Arh;
end;

Procedure TFoxDSet.MoveRecs(source,tagert:integer);
var RecSource,bubu1:Pointer;
    m,i:integer;
begin
  getmem(RecSource,FHeader.HeaderLen);
  getmem(bubu1,FHeader.HeaderLen);
  try
     if tagert>source then i:=1 else i:=-1; // direct
    DBFile.Seek((FHeader.HeaderLen) + ((source - 1) * FHeader.RecordLen), soFromBeginning);
    DBFile.ReadBuffer(RecSource^, FHeader.RecordLen);
    m:=source;
     while m<>tagert do begin
     m := m + i;
     DBFile.Seek((FHeader.HeaderLen) + ((m - 1) * FHeader.RecordLen), soFromBeginning);
     DBFile.ReadBuffer(bubu1^, FHeader.RecordLen);
     DBFile.Seek((FHeader.HeaderLen) + ((m - 1 - i) * FHeader.RecordLen), soFromBeginning);
     DBFile.WriteBuffer(bubu1^, FHeader.RecordLen);
     end;
    DBFile.Seek((FHeader.HeaderLen) + ((tagert - 1) * FHeader.RecordLen), soFromBeginning);
    DBFile.WriteBuffer(RecSource^, FHeader.RecordLen);
  finally
  freemem(bubu1);
  freemem(RecSource);
  end;
end;

function TFoxDSet.Locate(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions): Boolean;
var   Fields: TList;
           i: integer;
           b: boolean;
begin
  CheckBrowseMode;
  Fields := TList.Create;
    GetFieldList(Fields, KeyFields);
    First;
            b := False;
try
    While not eof do begin
            b := True;
            For i:=0 to Fields.count-1 do
            begin
             if VarIsArray(KeyValues) then
             b := KeyValues[i] = TField(Fields[i]).Value else
             b := KeyValues = TField(Fields[i]).Value;
            if not b then break;
            end;
    if b then break;
    next;
    end;
finally
Result := b;
end;
end;

function TFoxDSet.Lookup(const KeyFields: string; const KeyValues: Variant;
  const ResultFields: string): Variant;
begin
  Result := Null;
  if Locate(KeyFields, KeyValues, []) then
  begin
    SetTempState(dsCalcFields);
    try
      CalculateFields(TempBuffer);
      RestoreState(dsBrowse);
      Result := FieldValues[ResultFields];
    finally
    end;
  end;
end;

{ TMemoFieldP }
constructor TMemoFieldP.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetDataType(ftMemo);
  Transliterate := True;
  FieldKnow := False;
end;

procedure TMemoFieldP.Clear;
begin
{No clear to file}
end;

procedure TMemoFieldP.Assign(Source: TPersistent);
begin
  if (Source is TPicture) or (Source is TBitmap)
      or (Source is TMetaFile)  or (Source is TIcon) then
  begin
    LoadFromPicture(Source);
    Exit;
  end;
  inherited AssignTo(Source);
end;

Procedure TMemoFieldP.LoadFromPicture(Source: TPersistent);
var
  BlobStream: TStream;
  Picture: TGraphic;
begin
//  TypeGraph := 0;     {for Write - NoConvert}
//  FieldKnow := True;

  BlobStream := DataSet.CreateBlobStream(Self, bmWrite);
  Transliterate := False;
try
  if Source is TPicture then Picture := TPicture(Source).Graphic
                        else Picture := TGraphic(Source);
  Picture.SaveToStream(BlobStream);
finally
BlobStream.Free;
end;
end;

procedure TMemoFieldP.AssignTo(Dest: TPersistent);
begin
  if (Dest is TPicture) or (Dest is TBitmap)
      or (Dest is TMetaFile)  or (Dest is TIcon) then
  begin
    SaveToPicture(Dest);
    Exit;
  end;
  inherited AssignTo(Dest);
end;

Procedure TMemoFieldP.SaveToPicture(Source: TPersistent);
var
  BlobStream: TStream;
  Picture: TGraphic;
begin
  BlobStream := DataSet.CreateBlobStream(Self, bmRead);
  Transliterate := False;
Picture := nil;
try
if FieldKnow then begin
  case TypeGraph of
  $4D42:{bmp} Picture := TBitMap.Create;
  $D8FF:{jpg} Picture := TJPEGImage.Create;
  $CDD7:{wmf} Picture := TMetaFile.Create;
//  $0001:{emf} TMetaFile(Picture).LoadFromStream(BlobStream);
  $0000:{ico} Picture := TIcon.Create;
  end;
  if Picture <> nil then Picture.LoadFromStream(BlobStream);
  if Source is TPicture then TPicture(Source).Graphic := Picture
                        else TGraphic(Source) := Picture;

end;
finally
Picture.Free;
BlobStream.Free;
end;
end;

function TFoxDSet.Translate(Src, Dest: PChar; ToOem: Boolean): Integer;
begin
  Result := StrLen(Src);
  if FTranlateOn then begin // if copy no translate
  If assigned(FOnConvert) then FOnConvert(Src, Dest, ToOem)
  else
  if ToOem then
    CharToOemBuff(Src,Dest,Result) else
    OemToCharBuff(Src,Dest,Result);
  if Src <> Dest then Dest[Result] := #0;
                      end;
end;

Procedure TFoxDSet.UpdateFieldData(FieldNo: integer; RecBuf: Pointer; Buffer: Pointer);
Var
  Res: PChar;
  PointSet: Pointer;
  i,j,count:integer;
  F :PDouble;
  L :PInt64;
  Ismoll:Pinteger;
  TS: TTimeStamp;
Begin
If (FieldNo >= 0) And (FieldNo <= FNFields) Then Begin
Count := FieldStruct[FieldNo].FieldLength;
PointSet := Buffer; // Default
case FieldStruct[FieldNo].FieldType of
'C':    begin
           j := Length(PChar(Buffer));
           i := Count - j;
           if i>0 then FillChar(PChar(Buffer)[j],i,' '); // !!! if buffer = MAXstring
        end;
'N','F':begin
        Res := Pchar(FloatToStrF(Double(Buffer^),ffFixed,
                        Count,
                        FieldStruct[FieldNo].Decimals));
        j := Length(Res);
        i := Count - j;
        if i>0 then Res := Pchar(StringOfChar(' ',i)+Res);
        PointSet := Res;
        end;
'Y':    begin
        F := Buffer;
        L := Buffer;
        L^ := round(F^ * 10000);  {PointSet changed automatic}
        end;
'T':    begin
        F := Buffer;
        TS := DateTimeToTimeStamp(F^);
        Ismoll := Buffer;
        Ismoll^ := TS.Date+1721425;
        inc(Ismoll,1);
        Ismoll^ := TS.Time;
           end;
'L':    begin
        If Byte(Buffer^) > 0 then PChar(Buffer)[0]:='T'
        else PChar(Buffer)[0]:='F';
        end;
'D','B','I','G':;
end;
        Move(PointSet^, PChar(RecBuf)[FieldStruct[FieldNo].Address-1], Count);
  End;
End;

function TFoxDSet.GetFieldS(FieldD: TFieldDef; var Fields:TFieldStruct):boolean;
begin
    Result := True;
    FillChar(FieldS,SizeOf(FieldS),0);
    FieldD.Name := UpperCase(FieldD.Name);
    Move(PChar(FieldD.Name)^,FieldS.FieldName,Length(FieldD.Name));
    FieldS.FieldLength := FieldD.Size;
    FieldS.Decimals := FieldD.Precision;
    case FieldD.DataType of
      ftString: FieldS.FieldType := 'C';
      ftCurrency: begin FieldS.FieldType := 'Y';
                        FieldS.FieldLength := 8;  end;
      ftDate: begin   FieldS.FieldType := 'D';
                      FieldS.FieldLength := 8;  end;
      ftDateTime: begin FieldS.FieldType := 'T';
                        FieldS.Reserved := 4; {bin}
                        FieldS.FieldLength := 8;  end;
      ftBoolean: begin FieldS.FieldType := 'L';
                       FieldS.FieldLength := 1;  end;
      ftFloat: if TableType = Fox30 then begin
                     FieldS.FieldType := 'B';
                     FieldS.Reserved := 4; {bin}
                     FieldS.FieldLength := 8;  end
                else begin FieldS.FieldType := 'N';
                     if FieldS.FieldLength =0 then FieldS.FieldLength := 8;
                     end;
      ftInteger: begin FieldS.FieldType := 'I';
                       FieldS.Reserved := 4; {bin}
                       FieldS.FieldLength := 4;
                 end;
      ftDBaseOle: begin FieldS.FieldType := 'G';
                        FieldS.FieldLength := 4;
                 end;
      ftMemo: begin FieldS.FieldType := 'M';
              if TableType = Fox30 then FieldS.FieldLength := 4
                                   else FieldS.FieldLength := 10;
              end;
      else Result := false;
      end;
end;

function TFoxDSet.GetTableType: TTypeDBF;
begin
case FTypeDBF of
        $03: Result := dB3;
        $04: Result := dB4;
        $05: Result := dB5;
        $F5: Result := Fox2x;
        $30: Result := Fox30;
        $8B: Result := dBxMemo;
        else Result := UnKnown;
end;
end;

procedure TFoxDSet.SetTableType(Value: TTypeDBF);
begin
case Value of
        dB3: FTypeDBF := $03;
        dB4: FTypeDBF := $04;
        dB5: FTypeDBF := $05;
      dBxMemo: FTypeDBF := $8B;
        Fox2x: FTypeDBF := $F5;
        Fox30: FTypeDBF := $30;
        else FTypeDBF := 0;
end;
end;

function TFoxDSet.SizeDataFile:integer;
begin
  Result :=0;
  if (FIsOpen) then Begin
  Result := DBFile.Size;
  if MemoFile <> nil then
     Result := Result + MemoFile.Size;
                    End;
end;

end.
